/**********************************************************************
Copyright (c) 2012 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.datanucleus.store.neo4j;

import java.sql.SQLException;
import java.util.Map;

import javax.sql.XAConnection;
import javax.transaction.xa.XAException;
import javax.transaction.xa.XAResource;
import javax.transaction.xa.Xid;

import org.datanucleus.exceptions.NucleusDataStoreException;
import org.datanucleus.exceptions.NucleusException;
import org.datanucleus.store.StoreManager;
import org.datanucleus.store.connection.AbstractConnectionFactory;
import org.datanucleus.store.connection.AbstractManagedConnection;
import org.datanucleus.store.connection.ManagedConnection;
import org.datanucleus.store.connection.ManagedConnectionResourceListener;
import org.datanucleus.util.Localiser;
import org.datanucleus.util.NucleusLogger;
import org.neo4j.graphdb.GraphDatabaseService;
import org.neo4j.graphdb.Transaction;
import org.neo4j.graphdb.factory.GraphDatabaseFactory;

/**
 * Implementation of a ConnectionFactory for Neo4j.
 * Accepts a URL of the form 
 * <pre>neo4j:{db_path}</pre>
 * If {db_path} is not specified then will use "datanucleus" as the DB_PATH.
 */
public class ConnectionFactoryImpl extends AbstractConnectionFactory
{
    protected static final Localiser LOCALISER = Localiser.getInstance(
        "org.datanucleus.store.mongodb.Localisation", Neo4jStoreManager.class.getClassLoader());

    String dbName = "DataNucleus";

    GraphDatabaseFactory factory;

    /**
     * Constructor.
     * @param storeMgr Store Manager
     * @param resourceType Type of resource (tx, nontx)
     */
    public ConnectionFactoryImpl(StoreManager storeMgr, String resourceType)
    {
        super(storeMgr, resourceType);

        // "neo4j:[db_path]"
        String url = storeMgr.getConnectionURL();
        if (url == null)
        {
            throw new NucleusException("You haven't specified persistence property 'datanucleus.ConnectionURL' (or alias)");
        }

        String remains = url.substring(6).trim();
        if (remains.indexOf(':') == 0)
        {
            remains = remains.substring(1);
        }

        // Assumed to be DB_PATH
        if (remains.length() > 0)
        {
            dbName = remains;
        }

        factory = new GraphDatabaseFactory();
    }

    public void close()
    {
        super.close();
    }

    /**
     * Obtain a connection from the Factory. The connection will be enlisted within the {@link org.datanucleus.Transaction} 
     * associated to the <code>poolKey</code> if "enlist" is set to true.
     * @param poolKey the pool that is bound the connection during its lifecycle (or null)
     * @param transactionOptions Any options for then creating the connection
     * @return the {@link org.datanucleus.store.connection.ManagedConnection}
     */
    public ManagedConnection createManagedConnection(Object poolKey, Map transactionOptions)
    {
        return new ManagedConnectionImpl(transactionOptions);
    }

    public class ManagedConnectionImpl extends AbstractManagedConnection
    {
        Map options;
        Transaction graphTx;
        boolean committed = false;

        public ManagedConnectionImpl(Map options)
        {
            this.options = options;
        }

        public Object getConnection()
        {
            if (conn == null)
            {
                // Get the graphDB connection
                GraphDatabaseService graphDB = factory.newEmbeddedDatabase(dbName);
                registerShutdownHook(graphDB);
                conn = graphDB;
                graphTx = graphDB.beginTx();
            }
            committed = false;
            return conn;
        }

        public void release()
        {
            if (commitOnRelease)
            {
                if (conn != null)
                {
                    graphTx.success();
                    graphTx.finish();
                    committed = true;
                }
            }
            super.release();
        }

        public void close()
        {
            if (conn == null)
            {
                return;
            }
            for (int i=0; i<listeners.size(); i++)
            {
                ((ManagedConnectionResourceListener)listeners.get(i)).managedConnectionPreClose();
            }

            // TODO If running via a transaction (EmulatedXAResource) we may already have done the commit so should omit
            if (!committed)
            {
                graphTx.success();
                graphTx.finish();
            }

            // Shutdown the graphDB
            ((GraphDatabaseService)conn).shutdown();

            for (int i=0; i<listeners.size(); i++)
            {
                ((ManagedConnectionResourceListener)listeners.get(i)).managedConnectionPostClose();
            }

            conn = null;
            graphTx = null;
        }

        public XAResource getXAResource()
        {
            if (getConnection() instanceof GraphDatabaseService)
            {
                return new EmulatedXAResource(graphTx);
            }
            else
            {
                try
                {
                    return ((XAConnection)getConnection()).getXAResource();
                }
                catch (SQLException e)
                {
                    throw new NucleusDataStoreException(e.getMessage(),e);
                }
            }
        }

        private void registerShutdownHook(final GraphDatabaseService graphDb)
        {
            Runtime.getRuntime().addShutdownHook(new Thread()
            {
                @Override
                public void run()
                {
                    graphDb.shutdown();
                }
            });
        }
    }

    /**
     * Emulate the two phase protocol for non XA
     */
    static class EmulatedXAResource implements XAResource
    {
        Transaction graphTx;

        EmulatedXAResource(Transaction graphTx)
        {
            this.graphTx = graphTx;
        }

        public void start(Xid xid, int flags) throws XAException
        {
            NucleusLogger.CONNECTION.debug("Managed connection "+this.toString()+
                " is starting for transaction "+xid.toString()+" with flags "+flags);
        }

        public void commit(Xid xid, boolean onePhase) throws XAException
        {
            NucleusLogger.CONNECTION.debug("Managed connection "+this.toString()+
                " is committing for transaction "+xid.toString()+" with onePhase="+onePhase);

            graphTx.success();
            graphTx.finish();
            NucleusLogger.CONNECTION.debug("Managed connection "+this.toString()+
                " committed connection for transaction "+xid.toString()+" with onePhase="+onePhase);
        }

        public void rollback(Xid xid) throws XAException
        {
            NucleusLogger.CONNECTION.debug("Managed connection "+this.toString()+
                " is rolling back for transaction "+xid.toString());

            graphTx.failure();
            graphTx.finish();
            NucleusLogger.CONNECTION.debug("Managed connection "+this.toString()+
                " rolled back connection for transaction "+xid.toString());
        }

        public void end(Xid xid, int flags) throws XAException
        {
            NucleusLogger.CONNECTION.debug("Managed connection "+this.toString()+
                " is ending for transaction "+xid.toString()+" with flags "+flags);
            //ignore
        }

        public void forget(Xid xid) throws XAException
        {
            //ignore
        }

        public int prepare(Xid xid) throws XAException
        {
            NucleusLogger.CONNECTION.debug("Managed connection "+this.toString()+
                " is preparing for transaction "+xid.toString());
            return 0;
        }

        public Xid[] recover(int flags) throws XAException
        {
            throw new XAException("Unsupported operation");
        }

        public int getTransactionTimeout() throws XAException
        {
            return 0;
        }

        public boolean setTransactionTimeout(int timeout) throws XAException
        {
            return false;
        }

        public boolean isSameRM(XAResource xares) throws XAException
        {
            return (this == xares);
        }
    }
}