/**********************************************************************
Copyright (c) 2012 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.datanucleus.store.neo4j.fieldmanager;

import java.lang.reflect.Array;
import java.util.Collection;
import java.util.Iterator;

import org.datanucleus.ClassLoaderResolver;
import org.datanucleus.exceptions.NucleusDataStoreException;
import org.datanucleus.exceptions.NucleusException;
import org.datanucleus.exceptions.NucleusUserException;
import org.datanucleus.metadata.AbstractClassMetaData;
import org.datanucleus.metadata.AbstractMemberMetaData;
import org.datanucleus.metadata.FieldPersistenceModifier;
import org.datanucleus.metadata.FieldRole;
import org.datanucleus.metadata.MetaDataUtils;
import org.datanucleus.metadata.Relation;
import org.datanucleus.store.ExecutionContext;
import org.datanucleus.store.ObjectProvider;
import org.datanucleus.store.fieldmanager.AbstractFieldManager;
import org.datanucleus.store.fieldmanager.FieldManager;
import org.datanucleus.store.neo4j.Neo4jStoreManager;
import org.datanucleus.store.neo4j.Neo4jUtils;
import org.datanucleus.store.schema.naming.ColumnType;
import org.datanucleus.store.types.converters.TypeConverter;
import org.datanucleus.store.types.sco.SCOUtils;
import org.neo4j.graphdb.Node;
import org.neo4j.graphdb.Relationship;
import org.neo4j.graphdb.RelationshipType;

/**
 * Field Manager for retrieving values from Neo4j.
 */
public class FetchFieldManager extends AbstractFieldManager
{
    protected ExecutionContext ec;

    protected ObjectProvider op;

    protected Node node;

    protected AbstractClassMetaData cmd;

    boolean embedded = false;

    /** Metadata for the owner field if this is embedded. */
    protected AbstractMemberMetaData ownerMmd = null;

    public FetchFieldManager(ObjectProvider op, Node node)
    {
        this(op.getExecutionContext(), node, op.getClassMetaData());
        this.op = op;
        if (op.getEmbeddedOwners() != null)
        {
            embedded = true;
        }
    }

    public FetchFieldManager(ExecutionContext ec, Node node, AbstractClassMetaData acmd)
    {
        this.ec = ec;
        this.node = node;
        this.cmd = acmd;
        if (node == null)
        {
            throw new NucleusException("Attempt to create FetchFieldManager for " + op + " with null Neo4j Node!" +
                " Generate a testcase that reprocues this and raise an issue");
        }
    }

    protected String getPropName(int fieldNumber)
    {
        return ec.getStoreManager().getNamingFactory().getColumnName(
            cmd.getMetaDataForManagedMemberAtAbsolutePosition(fieldNumber), ColumnType.COLUMN);
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchBooleanField(int)
     */
    @Override
    public boolean fetchBooleanField(int fieldNumber)
    {
        return (Boolean)node.getProperty(getPropName(fieldNumber));
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchByteField(int)
     */
    @Override
    public byte fetchByteField(int fieldNumber)
    {
        return (Byte)node.getProperty(getPropName(fieldNumber));
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchCharField(int)
     */
    @Override
    public char fetchCharField(int fieldNumber)
    {
        return (Character)node.getProperty(getPropName(fieldNumber));
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchDoubleField(int)
     */
    @Override
    public double fetchDoubleField(int fieldNumber)
    {
        return (Double)node.getProperty(getPropName(fieldNumber));
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchFloatField(int)
     */
    @Override
    public float fetchFloatField(int fieldNumber)
    {
        return (Float)node.getProperty(getPropName(fieldNumber));
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchIntField(int)
     */
    @Override
    public int fetchIntField(int fieldNumber)
    {
        return (Integer)node.getProperty(getPropName(fieldNumber));
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchLongField(int)
     */
    @Override
    public long fetchLongField(int fieldNumber)
    {
        return (Long)node.getProperty(getPropName(fieldNumber));
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchShortField(int)
     */
    @Override
    public short fetchShortField(int fieldNumber)
    {
        return (Short)node.getProperty(getPropName(fieldNumber));
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchStringField(int)
     */
    @Override
    public String fetchStringField(int fieldNumber)
    {
        return (String)node.getProperty(getPropName(fieldNumber));
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchObjectField(int)
     */
    @Override
    public Object fetchObjectField(int fieldNumber)
    {
        AbstractMemberMetaData mmd = cmd.getMetaDataForManagedMemberAtAbsolutePosition(fieldNumber);
        if (mmd.getPersistenceModifier() != FieldPersistenceModifier.PERSISTENT)
        {
            return op.provideField(fieldNumber);
        }

        ClassLoaderResolver clr = ec.getClassLoaderResolver();
        int relationType = mmd.getRelationType(clr);

        boolean embedded = false;
        if (this.embedded || ownerMmd != null)
        {
            embedded = true;
        }
        else if (relationType != Relation.NONE)
        {
            if (mmd.isEmbedded())
            {
                embedded = true;
            }
            else if (mmd.getEmbeddedMetaData() != null)
            {
                embedded = true;
            }
            else if (Relation.isRelationMultiValued(relationType))
            {
                if (mmd.hasCollection() && mmd.getElementMetaData() != null && mmd.getElementMetaData().getEmbeddedMetaData() != null)
                {
                    // Embedded collection element
                    embedded = true;
                }
                else if (mmd.hasArray() && mmd.getElementMetaData() != null && mmd.getElementMetaData().getEmbeddedMetaData() != null)
                {
                    // Embedded collection element
                    embedded = true;
                }
                else if (mmd.hasMap() && 
                        ((mmd.getKeyMetaData() != null && mmd.getKeyMetaData().getEmbeddedMetaData() != null) || 
                        (mmd.getValueMetaData() != null && mmd.getValueMetaData().getEmbeddedMetaData() != null)))
                {
                    // Embedded map key/value
                    embedded = true;
                }
            }
        }

        if (embedded)
        {
            if (Relation.isRelationSingleValued(relationType))
            {
                // Embedded PC object
                // TODO Detect null embedded object
                if (ownerMmd != null)
                {
                    // Detect bidirectional relation so we know when to stop embedding
                    if (Relation.isBidirectional(relationType))
                    {
                        if ((ownerMmd.getMappedBy() != null && mmd.getName().equals(ownerMmd.getMappedBy())) ||
                            (mmd.getMappedBy() != null && ownerMmd.getName().equals(mmd.getMappedBy())))
                        {
                            // Other side of owner bidirectional, so return the owner
                            ObjectProvider[] ownerOps = op.getEmbeddedOwners();
                            return (ownerOps != null && ownerOps.length > 0 ? ownerOps[0].getObject() : null);
                        }
                    }
                    else
                    {
                        // mapped-by not set but could have owner-field
                        if (ownerMmd.getEmbeddedMetaData() != null &&
                            ownerMmd.getEmbeddedMetaData().getOwnerMember() != null &&
                            ownerMmd.getEmbeddedMetaData().getOwnerMember().equals(mmd.getName()))
                        {
                            // This is the owner-field linking back to the owning object so return the owner
                            ObjectProvider[] ownerOps = op.getEmbeddedOwners();
                            return (ownerOps != null && ownerOps.length > 0 ? ownerOps[0].getObject() : null);
                        }
                    }
                }

                AbstractClassMetaData embcmd = ec.getMetaDataManager().getMetaDataForClass(mmd.getType(), clr);
                if (embcmd == null)
                {
                    throw new NucleusUserException("Field " + mmd.getFullFieldName() +
                        " marked as embedded but no such metadata");
                }

                // Extract the owner member metadata for this embedded object
                AbstractMemberMetaData embMmd = mmd;
                if (ownerMmd != null)
                {
                    // Nested, so use from the embeddedMetaData
                    embMmd = ownerMmd.getEmbeddedMetaData().getMemberMetaData()[fieldNumber];
                }

                // TODO Cater for inherited embedded objects (discriminator)

                ObjectProvider embOP = ec.newObjectProviderForEmbedded(mmd, embcmd, op, fieldNumber);
                FieldManager ffm = new FetchEmbeddedFieldManager(embOP, node, embMmd);
                embOP.replaceFields(embcmd.getAllMemberPositions(), ffm);
                return embOP.getObject();
            }
            throw new NucleusUserException("Dont currently support embedded field : " + mmd.getFullFieldName());
        }

        if (Relation.isRelationSingleValued(relationType))
        {
            RelationshipType type = DNRelationshipType.SINGLE_VALUED;
            if (relationType == Relation.MANY_TO_ONE_BI)
            {
                type = DNRelationshipType.MULTI_VALUED;
            }

            AbstractClassMetaData relCmd = null;
            String requiredFieldName = mmd.getName();
            if (mmd.getMappedBy() != null)
            {
                AbstractMemberMetaData[] relMmds = mmd.getRelatedMemberMetaData(clr);
                requiredFieldName = relMmds[0].getName();
                relCmd = relMmds[0].getAbstractClassMetaData();
            }
            else if (relationType == Relation.MANY_TO_ONE_BI)
            {
                AbstractMemberMetaData[] relMmds = mmd.getRelatedMemberMetaData(clr);
                requiredFieldName = relMmds[0].getName();
                relCmd = relMmds[0].getAbstractClassMetaData();
            }
            else
            {
                relCmd = ec.getMetaDataManager().getMetaDataForClass(mmd.getType(), clr);
            }

            Iterable<Relationship> rels = node.getRelationships(type);
            if (rels != null)
            {
                Iterator<Relationship> relIter = rels.iterator();
                while (relIter.hasNext())
                {
                    Relationship rel = relIter.next();
                    String fieldName = (String) rel.getProperty(Neo4jStoreManager.RELATIONSHIP_FIELD_NAME);
                    if (fieldName != null && fieldName.equals(requiredFieldName))
                    {
                        Node otherNode = rel.getOtherNode(node);
                        return Neo4jUtils.getObjectForNode(otherNode, relCmd, ec, false, null);
                    }
                }
            }
            return null;
        }
        else if (Relation.isRelationMultiValued(relationType))
        {
            if (mmd.hasCollection())
            {
                Collection<Object> coll;
                try
                {
                    Class instanceType = SCOUtils.getContainerInstanceType(mmd.getType(), mmd.getOrderMetaData() != null);
                    coll = (Collection<Object>) instanceType.newInstance();
                }
                catch (Exception e)
                {
                    throw new NucleusDataStoreException(e.getMessage(), e);
                }

                AbstractClassMetaData elemCmd = mmd.getCollection().getElementClassMetaData(clr, ec.getMetaDataManager());
                if (elemCmd == null)
                {
                    // Try any listed implementations
                    String[] implNames = MetaDataUtils.getInstance().getImplementationNamesForReferenceField(mmd, 
                        FieldRole.ROLE_COLLECTION_ELEMENT, clr, ec.getMetaDataManager());
                    if (implNames != null && implNames.length == 1)
                    {
                        elemCmd = ec.getMetaDataManager().getMetaDataForClass(implNames[0], clr);
                    }
                    if (elemCmd == null)
                    {
                        throw new NucleusUserException("We do not currently support the field type of " + mmd.getFullFieldName() +
                            " which has a collection of interdeterminate element type (e.g interface or Object element types)");
                    }
                }

                Iterable<Relationship> rels = node.getRelationships(DNRelationshipType.MULTI_VALUED);
                if (rels != null)
                {
                    Iterator<Relationship> relIter = rels.iterator();
                    while (relIter.hasNext())
                    {
                        Relationship rel = relIter.next();
                        String fieldName = (String) rel.getProperty(Neo4jStoreManager.RELATIONSHIP_FIELD_NAME);
                        if (fieldName != null && fieldName.equals(mmd.getName()))
                        {
                            if (rel.hasProperty(Neo4jStoreManager.RELATIONSHIP_INDEX_NAME))
                            {
                                // TODO Make use of "index" if handling a list
                            }
                            Node elemNode = rel.getOtherNode(node);
                            Object elemPC = Neo4jUtils.getObjectForNode(elemNode, elemCmd, ec, false, null);
                            coll.add(elemPC);
                        }
                    }
                }
                return coll;
            }
            else if (mmd.hasArray())
            {
                AbstractClassMetaData elemCmd = mmd.getArray().getElementClassMetaData(clr, ec.getMetaDataManager());
                if (elemCmd == null)
                {
                    // Try any listed implementations
                    String[] implNames = MetaDataUtils.getInstance().getImplementationNamesForReferenceField(mmd, 
                        FieldRole.ROLE_ARRAY_ELEMENT, clr, ec.getMetaDataManager());
                    if (implNames != null && implNames.length == 1)
                    {
                        elemCmd = ec.getMetaDataManager().getMetaDataForClass(implNames[0], clr);
                    }
                    if (elemCmd == null)
                    {
                        throw new NucleusUserException("We do not currently support the field type of " + mmd.getFullFieldName() +
                            " which has an array of interdeterminate element type (e.g interface or Object element types)");
                    }
                }

                Object array = null;
                int arraySize = 0;
                Iterable<Relationship> rels = node.getRelationships(DNRelationshipType.MULTI_VALUED);
                if (rels != null)
                {
                    Iterator<Relationship> relIter = rels.iterator();
                    while (relIter.hasNext())
                    {
                        Relationship rel = relIter.next();
                        String fieldName = (String) rel.getProperty(Neo4jStoreManager.RELATIONSHIP_FIELD_NAME);
                        if (fieldName != null && fieldName.equals(mmd.getName()))
                        {
                            arraySize++;
                        }
                    }

                    int i = 0;
                    array = Array.newInstance(mmd.getType().getComponentType(), arraySize);

                    relIter = rels.iterator();
                    while (relIter.hasNext())
                    {
                        Relationship rel = relIter.next();
                        String fieldName = (String) rel.getProperty(Neo4jStoreManager.RELATIONSHIP_FIELD_NAME);
                        if (fieldName != null && fieldName.equals(mmd.getName()))
                        {
                            int position = i;
                            if (rel.hasProperty(Neo4jStoreManager.RELATIONSHIP_INDEX_NAME))
                            {
                                position = (Integer) rel.getProperty(Neo4jStoreManager.RELATIONSHIP_INDEX_NAME);
                            }
                            Node elemNode = rel.getOtherNode(node);
                            Object elemPC = Neo4jUtils.getObjectForNode(elemNode, elemCmd, ec, false, null);
                            Array.set(array, position, elemPC);
                            i++;
                        }
                    }
                }

                return array;
            }
            else if (mmd.hasMap())
            {
                // TODO Support this
                throw new NucleusUserException("Don't currently support 1-N/M-N maps : " + mmd.getFullFieldName());
            }
        }

        String fieldName = getPropName(fieldNumber);
        if (!node.hasProperty(fieldName))
        {
            return null;
        }
        Object value = node.getProperty(fieldName);

        if (mmd.isSerialized())
        {
            throw new NucleusUserException("Dont currently support serialised fields : " + mmd.getFullFieldName());
        }

        if (mmd.getTypeConverterName() != null)
        {
            // User-defined type converter
            TypeConverter conv = ec.getNucleusContext().getTypeManager().getTypeConverterForName(mmd.getTypeConverterName());
            return conv.toMemberType(value);
        }

        Object fieldValue = Neo4jUtils.getFieldValueFromStored(ec, mmd, value, FieldRole.ROLE_FIELD);
        if (op != null)
        {
            // Wrap if SCO
            return op.wrapSCOField(mmd.getAbsoluteFieldNumber(), fieldValue, false, false, true);
        }
        return fieldValue;
    }
}